/*
   fuse - associative image reconstruction
   Copyright (C) 1997  Scott Draves <spot@cs.cmu.edu>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define argi(s,d)   ((ai = getenv(s)) ? atoi(ai) : (d))
#define argd(s,d)   ((ai = getenv(s)) ? atof(ai) : (d))
#define args(s,d)   ((ai = getenv(s)) ? ai : (d))

#define USE_GL

#ifdef USE_GL
#include <gl.h>
#include "device.h"
#endif

#include <math.h>
#include "image.h"

int size, overlap;
int ntries;
double ratio;

/* find difference between two rects.  image1
   is assumed to be the "mask", ie if its pixel
   value is 0, then that pixel doesn't matter
   */
int
compare(image1, x1, y1, image2, x2, y2, nmasked, w, h)
     image *image1, *image2;
     int x1, y1, x2, y2, *nmasked, w, h;
{
  int t, i, j, diff = 0;
  int stride1 = image1->width;
  int stride2 = image2->width;

  CLIP(image1, x1, y1, w, h);
  CLIP(image2, x2, y2, w, h);

  *nmasked = 0;
  for (i = 0; i < h; i++)
    for (j =0; j < w; j++) {
      uint p1 = image1->pixels[(y1 + i) * stride1 + x1 + j];
      uint p2 = image2->pixels[(y2 + i) * stride2 + x2 + j];
      if (p1) {
	int k, shift;
	uint mask;
	shift = 0;
	mask = 0xff;
	for (k = 0; k < 4; k++) {
	  int t = (((p1 & mask) >> shift) -
		   ((p2 & mask) >> shift));
	  diff += t * t;
	  mask = mask << 8;
	  shift += 8;
	}
      } else {
	(*nmasked)++;
      }
    }
  return diff;
}

void
source_match(image *over,
	     image *images[], int nimages,
	     image *image2,
	     int x, int y)
{
   int nmasked, w = size, h = size;
   int i, best_x, best_y, best_diff = 1<<30;
   int this_x, this_y, d;
   image *image1, *best_image;
   CLIP(image2, x, y, w, h);
   for (i = 0; i < ntries; i++) {
      image1 = images[random() % nimages];
      this_x = random() % (image1->width - w);
      this_y = random() % (image1->height - h);
      d = compare(image2, x, y, image1, this_x, this_y, &nmasked, w, h);
      if (over) {
	 d *= ratio * w * h / (w * h - nmasked);
	 d += compare(over, x, y, image1, this_x, this_y, &nmasked, w, h);
      }
      if (d < best_diff) {
	 best_diff = d;
	 best_x = this_x;
	 best_y = this_y;
	 best_image = image1;
	 blit(best_image, best_x, best_y, image2, x, y, w, h);
	 printf("%d\n", d);
#ifdef USE_GL
	 lrectwrite(0, 0, image2->width - 1, image2->height - 1,
		    image2->pixels);
#endif
      }
   }
}


main(int argc, char **argv)
{
   char *ai;
   char *getenv();
   image *image1, *images[100];
   image *over;
   image *image2;
   int i, nimages;

   if (getenv("over")) {
      over = read_image(getenv("over"), 0);
      image2 = new_image(over->width, over->height);
   } else {
      int sz = argi("size", 1000);
      over = 0;
      image2 = new_image(sz, sz);
   }
   ntries = argi("ntries", 10000);
   ratio = argd("ratio", 1.0);
   size = argi("block", 40);
   overlap = argi("overlap", 10);
   for (i = 1; i < argc; i++) {
      printf("reading %s...\n", argv[i]);
      images[i-1] = read_image(argv[i], 0);
   }
   nimages = argc-1;

   srandom(time(0));
#ifdef USE_GL
   prefsize(image2->width, image2->height);
   foreground();
   winopen("fuse");
   RGBmode();
   pixmode(PM_STRIDE, image2->width);
   gconfig();
#endif
   if (1) {
      /* standard good values are size 40 and overlap 10, though
	 this depends on the feature size of the source images and
	 * the style output you want.  generally, small sizes result in
	 * homogenious output */
      int r, i, parity = 0;
      int p[1000];
      int diag = sqrt(image2->width * image2->width +
		      image2->height * image2->height) / 2;
      image1 = images[0];
      blit(image1, (image1->width - size)/2, (image1->height - size)/2,
	   image2, (image2->width - size)/2, (image2->height - size)/2,
	   size, size);
      for (r = 1; r < diag; r += size - overlap) {
	 int n = 1 + r * 2.0 * M_PI / (size - overlap);
	 if (n > 1000)
	    printf("joiweugunvnsd\n");
	 for (i = 0; i < n; i++)
	    p[i] = i;
	 for (i = 0; i < 2*n; i++) {
	    int i1 = random() % n;
	    int i2 = random() % n;
	    int t = p[i1];
	    p[i1] = p[i2];
	    p[i2] = t;
	 }
	 parity++;
	 for (i = 0; i < n; i++) {
	    int x, y;
	    float theta = 2.0 * M_PI * p[i] / n;
	    if (parity&1)
	       theta += M_PI / n;
	    x = image2->width / 2 + r * cos(theta) - size/2;
	    y = image2->height / 2 + r * sin(theta) - size/2;
	    source_match(over, images, nimages, image2, x, y);
#ifdef USE_GL
	    lrectwrite(0, 0, image2->width - 1, image2->height - 1, image2->pixels);
#endif
	 }
      }
   }
   write_image(image2, args("out", "fuse.rgb"));
   gexit();
}
