/*
   fuse - associative image reconstruction
   Copyright (C) 1997  Scott Draves <spot@cs.cmu.edu>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <stdio.h>
#include <math.h>

#include "image.h"

#if 0
void
flip_coefs(img, x, y, w, h, rot,
	   base_r, kx_r, ky_r)
   image *img;
   int x, y, *w, *h, rot;
   uint *base_r;
   int *kx_r, *ky_r;
{
   uint base;
   int kx, ky;

   base = img->pixels + y * img->width + x;
   if (rot & 1) {
      base += *w;
      kx = -1;
   } else
      kx = 1;
   if (rot & 2) {
      base += *h * img->width;
      ky = -img->width;
   } else
      ky = img->width;
   if (rot & 4) {
      int t;
      t = kx;
      kx = ky;
      ky = t;
      t = *w;
      *w = *h;
      *h = t;
   }
   *kx_r = kx;
   *ky_r = ky;
   *base_r = base;
}
#endif

/* from image1 to image2 */
void
blit(image1, x1, y1, image2, x2, y2, w, h)
     image *image1, *image2;
     int x1, y1, x2, y2, w, h;
{
  int i, j;
  int w1, w2;
  uint *p1, *p2;
  CLIP(image1, x1, y1, w, h);
  CLIP(image2, x2, y2, w, h);
  w1 = image1->width;
  w2 = image2->width;
  p1 = image1->pixels;
  p2 = image2->pixels;
  for (i = 0; i < h; i++)
    for (j = 0; j < w; j++)
      p2[w2 * (y2 + i) + x2 + j] =
	p1[w1 * (y1 + i) + x1 + j];
}

/* sum of entries of vector to 1 */
void normalize_vector(v, n)
   double *v;
   int n;
{
   double t = 0.0;
   int i;
   for (i = 0; i < n; i++)
      t += v[i];
   t = 1.0 / t;
   for (i = 0; i < n; i++)
      v[i] *= t;
}

#define FILTER_CUTOFF 1

void
filter_down(image *source, image *dest)
{
  double *filter;
  uint *sp = source->pixels;
  uint *dp = dest->pixels;
  int sw = source->width;
  int dw = dest->width;
  int oversamp = sw / dw;
  int w, h, i, j, filter_width = 3 * oversamp;
  if (oversamp * dw != sw)
    fprintf(stderr, "filter_down only handles integral scaling\n");
  if ((filter_width ^ oversamp) & 1)
    filter_width++;
  filter = (double *) malloc(sizeof(double) * filter_width * filter_width);
  /* fill in the coefs */
  for (i = 0; i < filter_width; i++) {
    for (j = 0; j < filter_width; j++) {
      double ii = ((2.0 * i + 1.0) / filter_width - 1.0) * FILTER_CUTOFF;
      double jj = ((2.0 * j + 1.0) / filter_width - 1.0) * FILTER_CUTOFF;
      double t = exp(-2.0 * (ii * ii + jj * jj));
      filter[i + j * filter_width] = t;
    }
  }
  normalize_vector(filter, filter_width * filter_width);
#if 0
  for (i = 0; i < filter_width; i++) {
    for (j = 0; j < filter_width; j++) {
      printf("%5f ", filter[i + j * filter_width]);
    }
    printf("\n");
  }
#endif
  for (h = oversamp; h < dest->height - oversamp; h++)
    for (w = oversamp; w < dw - oversamp; w++) {
      double a[3];
      double b[3];
      double c;
      a[0] = a[1] = a[2] = 0.0;
      for (i = 0; i < filter_width; i++)
	for (j = 0; j < filter_width; j++) {
	  uint p = sp[(h * dw * oversamp + w) * oversamp + i * sw + j];
	  c = filter[i * filter_width + j];
	  FRAG(p, b);
	  a[0] += b[0] * c;
	  a[1] += b[1] * c;
	  a[2] += b[2] * c;
	}
      dp[h * dw + w] = ASSEMBLE((int)a[0], (int)a[1], (int)a[2]);
    }
  free(filter);
}
  

image *
read_image(char *filename, image *old) {
  FILE *f;
  image *r;
  int i, j, k, w, h, rez;

  f = fopen(filename, "r");
  if (NULL == f) {
    fprintf(stderr, "can't open %s.\n", filename);
    exit(1);
  }
  if (3 != fscanf(f, "P6 %d %d %d", &w, &h, &rez)) {
    fprintf(stderr, "bad format %s.\n", filename);
    exit(1);
  }
  if (255 != rez) {
    fprintf(stderr, "not resolution 255 %s.\n", filename);
    exit(1);
  }
  if (old &&
      old->width == w &&
      old->height == h) {
    r = old;
  } else {
    r = (image *) malloc(sizeof(image));
    r->width = w;
    r->height = h;
    r->pixels = (uint *) malloc(sizeof(uint) * w * r->height);
    if (NULL == r->pixels) {
      fprintf(stderr, "can't malloc.\n");
      exit(1);
    }
  }
  
  fread(r->pixels, 1, w * r->height * 3, f);

  /* insert gaps, be careful at last pixel */
  if (1) {
    char *p = (char *) r->pixels;
    int scan_ppm = w * h * 3;
    int scan_rgba = w * h * sizeof(uint);
    while (scan_ppm) {
      p[scan_rgba - 1] = 255;
      p[scan_rgba - 2] = p[scan_ppm - 3];
      p[scan_rgba - 3] = p[scan_ppm - 1];
      p[scan_rgba - 4] = p[scan_ppm - 2];
      scan_rgba -= 4;
      scan_ppm -= 3;
    }
  }

  fclose(f);

  return r;
}

void
write_image(image *im, char *filename) {
  FILE *f = fopen(filename, "w");
  int i, j, k;
  fprintf(f, "P6\n%d %d\n255\n", im->width, im->height);
  for (i = 0; i < im->height; i++)
    for (k = 0; k < im->width; k++)
      fwrite(im->pixels + i * im->width + k, 1, 3, f);
  fclose(f);
}

image *
new_image(int w, int h)
{
  image *i = (image *) malloc(sizeof(image));
  i->width = w;
  i->height = h;
  i->pixels = (uint *) malloc(sizeof(uint) * w * h);
  if (NULL == i->pixels) {
    fprintf(stderr, "can't malloc %d by %d image.\n", w, h);
    exit(1);
  }
  return i;
}

void
free_image(image *i)
{
  free(i->pixels);
  free(i);
}
